/*
	MMC/SD Card SPI mode Read/Write Mudule

	Copyright(C) 2002 Tatsuyuki Satoh
*/

/* #include <stdio.h> */
#include "mmc_if.h"
#include "mmc_spi.h"

/* Default Sector Size */
#define SECTOR_SIZE 512

/* Timeout for command response */
#define TIMEOUT_CMDRES 400
#define TIMEOUT_DATASTART 400
#define TIMEOUT_DATAEND   400

/* MMC command code maker */
#define MMC_CMD(n) (n|0x40)

/* MMC receive response */
static unsigned char get_response(int timeout)
{
	unsigned char res;
	do
	{
		res = mmc_serial_in();
		if(res!=0xff)
			return res;
	}while(--timeout);
	/* timeout */
	return MMC_TIMEOUT;
}

/* MMC command packet */
static unsigned char mmc_cmd(unsigned char cmd,unsigned long arg)
{
	int i;
	unsigned char res;

	/* dummy */
	mmc_serial_out(0xff);

	mmc_serial_out(cmd);     /* CMD  */
	mmc_serial_out(arg>>24); /* ARG1 */
	mmc_serial_out(arg>>16); /* ARG2 */
	mmc_serial_out(arg>>8);  /* ARG3 */
	mmc_serial_out(arg>>0);  /* ARG4 */
	mmc_serial_out(0x95);    /* CRC  */
	/* CMD Response */
	return get_response(TIMEOUT_CMDRES);
}

/* initialize MMC/SD card with SPI mode */
unsigned char mmc_init_spi(void)
{
	int i;
	unsigned char res;

	/* I/F Enable */
	if( !mmc_hw_start())
		return MMC_ERR_HW;
#if 0
	/* power off */
	mmc_hw_power(0);
	sleep(500);
#endif
	/* power on */
	mmc_hw_power(1);
	sleep(200);

	/* CS=H & 80clock dummy */
	mmc_hw_cs(1);
	for(i=0;i<10;i++)
		mmc_serial_out(0xff);

	/* CS=L & DUMMY 8 clock */
	mmc_hw_cs(0);
	mmc_serial_out(0xff);

	/* CMD0 (with CRC) Response = 0x01 */
	res = mmc_cmd( MMC_CMD(0) , 0);
	if(res != 0x01)
		return MMC_ERR_CMD0;

	/* CMD1 Response = 0x01 -> 0x00 */
	do
	{
		res = mmc_cmd( MMC_CMD(1) , 0);
		if(res == 0x00)
			return MMC_NO_ERR;
	}while(res==0x01);
	/* error */
	return MMC_ERR_CMD1;
}

/* Finish MMC/SD card access */
void mmc_disable(void)
{
	mmc_hw_stop();
}

/* SPI write sector */
unsigned char spi_write(unsigned long sector,unsigned char *data)
{
	int i;
	unsigned char res;

	/* sector to args(address) conversion */
	sector *= SECTOR_SIZE;

	/* command */
	res = mmc_cmd(MMC_CMD(24), sector);
	if(res!=0x00)
	{
/*		printf("CMD Response error %02X\n",res);*/
		return MMC_ERR_CMD;
	}

	/* dummy */
	mmc_serial_out(0xff);

	/* Data Token */
	mmc_serial_out(0xfe); /* StartByte */
	for(i = 0; i < SECTOR_SIZE ; i++)
		mmc_serial_out(data[i]); /* BlockData */
	mmc_serial_out(0);    /* CRC1 */
	mmc_serial_out(0);    /* CRC2 */

	/* data response */
	res = get_response(TIMEOUT_DATAEND);
	if( (res&0x1f) != 0x5)
		return MMC_ERR_DATA;

	/* wait for ready */
	do
	{
		res = mmc_serial_in();
/* printf("%02X:",res); */
	}
	while(res != 0xff);

	/* CS = HI */
	return MMC_NO_ERR;
}

/* SPI read sector */
unsigned char spi_read(unsigned long sector,unsigned char *data)
{
	int i;
	unsigned char res;

	/* sector to args(address) conversion */
	sector *= SECTOR_SIZE;

	/* command */
	res = mmc_cmd(MMC_CMD(17), sector);
	if(res!=0x00)
	{
/*		printf("CMD Response error %02X\n",res);*/
		return MMC_ERR_CMD;
	}

	/* wait for START BYTE or error byte */
	res = get_response(TIMEOUT_DATASTART);
	if(res!=0xfe)
	{
/*		printf("DataStart Error %02X",res);*/
		return MMC_ERR_DATA;
	}

	/* DATA BLOCK */
	for(i = 0; i < SECTOR_SIZE; i++)
		data[i] = mmc_serial_in();

	/* CRC */
	mmc_serial_in();
	mmc_serial_in();

	/* CS=H */
	return MMC_NO_ERR;
}

